/*
 * MP3 Player, Main Program, http://www.pjrc.com/tech/mp3
 * Copyright (c) 2000, PJRC.COM, LLC
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * As a specific exception to the GPL, the executable object code built
 * from this source code may be combined with hardware configuration data
 * files.  A hardware configuration data file is a set of data that is
 * transmitted to an intergrated circuit that is not a general purpose
 * microprocessor or microcontroller, in order to establish its normal
 * operation.  The process of combining the executable ojbect code built
 * from the GPL licensed source code with the hardware configuration data
 * shall be considered an aggregation of another work not based on the
 * Program.  While the GPL does not restrict use of the program, any
 * use restriction associated with the hardware configuration data (for
 * example, that it only be used with particular hardware) shall apply
 * to the combined file which includes a copy of the hardware configuration
 * data.
 *
 * Contact: paul@pjrc.com
 */



#include "paulmon2.h"
#include "as31glue.h"
#include "startup.h"
#include "printf.h"
#include "display.h"
#include "parse.h"
#include <8051.h>


#define PARM_ATTN 1
#define DEFAULT_ATTN 12
#define MAX_ATTN 64
#define PARM_RAND8_SEED 2
#define PARM_TREBLE 3
#define DEFAULT_TREBLE 0
#define PARM_BASS 4
#define DEFAULT_BASS 0


char is_state_normal(unsigned char state);
void main_directory_backup(char fd);
void read_directory(char dir, xdata struct dirent_struct *fileinfo);
char is_mp3_file(xdata struct dirent_struct *fileinfo);
unsigned char rand8(void);

#define S_READ_DIR              1
#define S_CACHING               2
#define S_PLAYING_AND_CACHING   3
#define S_PLAYING               4
#define S_FINISH_PLAYING        5
#define S_PAUSED                6


#define STA013_TREBLE 0x7B
#define STA013_BASS 0x7C
#define STA013_TONE_ATTEN 0x7D
#define STA013_MAX_TONE 12

bit dummy_bit;

volatile xdata unsigned char rand8_seed;
char rand_call_count;
unsigned int dir_position;
idata unsigned char ibuf[4];


xdata struct settings_struct {
	unsigned char attn;
	char treble;
	char bass;
	unsigned char tone_attn;
} setting;


void main()
{
	unsigned char i, event, num;
	char dir=-1, mp3_fd=-1;
	char r;
	char state=S_READ_DIR, prior_state=S_READ_DIR;
	char caching_step=0;
	xdata struct dirent_struct fileinfo;
	unsigned int block;
	//unsigned char attn;
	//char bass, treble;


	/* initialize everything */
	pm2_pstr(
"\r\n\r\n\
\\[\\@ 0PJRC MP3 Player 0.6.2.D\\]\r\n\
MP3 Player, Version 0.6.2.D, June 17, 2001\r\n\
Copyright (C) 2001, PJRC.COM, LLC\r\n\r\n\
Type 'QUIT' to return to the monitor\r\n\r\n");

	if (hardware_init() != 0) pm2_exit();
	pm2_pstr("Init interrupt serial I/O\r\n ");
	init_uart();	// from here forward, use printf, not pm2_pstr

	// this doesn't work yet
	timer_setup();

	printf("Test of interrupt driven serial I/O\r\n");

	printf("Waiting for Hard Drive to be ready\r\n");
	ide_init();
	//ide_wakeup();

	lcd_setup();
	parse_init();

	if (detect_filesystem() == 0) pm2_exit();
	print_memory_available();

	setting.attn = read_param_1byte(PARM_ATTN, DEFAULT_ATTN);
	sta013_set_attenuation(setting.attn);
	lcd_volume_set(setting.attn);

	setting.tone_attn = STA013_MAX_TONE;
	sta013_write(STA013_TONE_ATTEN, &setting.tone_attn);

	setting.treble = read_param_1byte(PARM_TREBLE, DEFAULT_TREBLE);
	sta013_write(STA013_TREBLE, &setting.treble);
	setting.bass = read_param_1byte(PARM_BASS, DEFAULT_BASS);
	sta013_write(STA013_BASS, &setting.bass);

	printf("\r\nMP3 Player, Begin Playback\r\n");

	debug = 0;

	/* read the root directory */
	dir = file_open_by_1st_cluster(get_root_1st_cluster());
	do_ide_loop(1);
	file_cache(dir, 0, 0xFFFFFFFF);	 // request entire directory cached
	do {
		do_ide_loop(4);
		r = file_cache_work(dir);
		if (r == 2) pm2_exit();
	} while (r == 0);
	do_ide_loop(8);
	file_seek(dir, 0);
	dir_position = 0;
	rand8_seed = read_param_1byte(PARM_RAND8_SEED, 0);
	rand_call_count = 10;	/* cause write to flash on first file played  */

#if 0
	/* testing dir_read */
	ide_sleep();
	do_ide_loop(8);
	debug = 1;
	do  {
		printf("\r\n");
		r = dir_read(dir, &fileinfo);
		printf("dir_read, r=%d\r\n", r);
		print_memory_available();
		printf("%s %s , clust=%ld, size=%ld (0x%lx)\r\n",
			&(*fileinfo.short_name),
			&(*fileinfo.long_name),
			fileinfo.cluster, fileinfo.size, fileinfo.size);

	} while (r == 0);
	printf("\r\n");
	print_memory_available();
	printf("\r\n");
	printf("\r\n");
	pm2_exit();
#endif

	lcd_display_filename("Warming up");

	/* this infinite loop runs all the actions of the MP3 player */

	while (1) {
		/* first, process any events from the user interface */
		parse_serial_input();
		event = get_next_event();
		switch (event) {
		  case E_NOTHING: {
			break;
		  }
		  case E_PLAY_PAUSE: {
			if (state == S_PAUSED) {
				printf("Play\r\n");
				state = prior_state;
				play_resume();
			} else {
				if (is_state_normal(state)) {
					printf("Pause\r\n");
					prior_state = state;
					state = S_PAUSED;
					play_suspend();
				} else {
					put_back_event(event);
				}
			}
			break;
		  }

		  case E_NEXT:
			if (is_state_normal(state)) {
				printf("Next File\r\n");
				play_abort();
				ide_flush();
				state = S_FINISH_PLAYING;
			} else {
				put_back_event(event);
			}
			break;
			
		  case E_PREV:
			if (is_state_normal(state)) {
				printf("Previous File\r\n");
				play_abort();
				ide_flush();
				if (rand8_seed == 0) {
					/* don't bother if in random mode */
					main_directory_backup(dir);
				}
				state = S_FINISH_PLAYING;
			} else {
				put_back_event(event);
			}
			break;

		  case E_RANDOM:
			if (rand8_seed) {
				printf("Random Off\r\n");
				rand8_seed = 0;
			} else {
				printf("Random On\r\n");
				rand8_seed = 0x5A;
			}
			ibuf[0] = rand8_seed;  /* wr parm needs idata ptr */
			lcd_rand_set(rand8_seed);
			write_flash_param(PARM_RAND8_SEED, ibuf);
			rand_call_count = 0;
			break;

		  case E_VOL_UP:
			printf("Vol Up\r\n");
			if (--setting.attn == 255) setting.attn = 0;
			sta013_set_attenuation(setting.attn);
			lcd_volume_set(setting.attn);
			// printf("attn set to %d\r\n", setting.attn);
			ibuf[0] = setting.attn;
			write_flash_param(PARM_ATTN, ibuf);
			break;

		  case E_VOL_DOWN:
			printf("Vol Down\r\n");
			if (++setting.attn > MAX_ATTN) setting.attn = MAX_ATTN;
			sta013_set_attenuation(setting.attn);
			lcd_volume_set(setting.attn);
			// printf("attn set to %d\r\n", setting.attn);
			ibuf[0] = setting.attn;
			write_flash_param(PARM_ATTN, ibuf);
			break;

		  case E_INC:
		  case E_DEC:
		  //case E_UP:
		  //case E_DOWN:
		  //case E_LEFT:
		  //case E_RIGHT:
		  case E_ENTER:
			lcd_user_action(event);
			break;

                  case E_UP:
                        printf("Bas Up, bass=%d\r\n", setting.bass);
                        if (++setting.bass > STA013_MAX_TONE)
				setting.bass = STA013_MAX_TONE;
                        sta013_write(STA013_BASS, &setting.bass);
                        lcd_bass_set(setting.bass);
                        break;

                  case E_DOWN:
                        printf("Bas Down, bass=%d\r\n", setting.bass);
                        if (--setting.bass < STA013_MAX_TONE * -1)
				setting.bass = STA013_MAX_TONE * -1;
                        sta013_write(STA013_BASS, &setting.bass);
                        lcd_bass_set(setting.bass);
                        break;

                  case E_RIGHT:
                        printf("Tre Up, treble=%d\r\n", setting.treble);
                        if (++setting.treble > STA013_MAX_TONE)
				setting.treble = STA013_MAX_TONE;
                        sta013_write(STA013_TREBLE, &setting.treble);
                        lcd_treble_set(setting.treble);
                        break;

                  case E_LEFT:
                        printf("Tre Down, treble=%d\r\n", setting.treble);
                        if (--setting.treble < STA013_MAX_TONE * -1)
				setting.treble = STA013_MAX_TONE * -1;
                        sta013_write(STA013_TREBLE, &setting.treble);
                        lcd_treble_set(setting.treble);
                        break;

		  case E_TIMER0:
		  case E_TIMER1:
		  case E_TIMER2:
		  case E_TIMER3:
			printf("Timer %d Event\r\n", event - E_TIMER0);
			break;

		  default:
			printf("\r\n\r\nUnsupported event #%d\r\n\r\n", event);
		}



		lcd_display_update();

		do_ide_loop(6);

		/* and then do the next step to actually play MP3 files */

		switch(state) {
		  case S_READ_DIR:
			read_directory(dir, &fileinfo);
			printf("Playing: %s %s , clust=%ld, size=%ld (0x%lx)\r\n",
				&(*fileinfo.short_name),
				&(*fileinfo.long_name),
				fileinfo.cluster, fileinfo.size, fileinfo.size);
			lcd_display_filename(&(*fileinfo.long_name));
			mp3_fd = file_open_by_1st_cluster(fileinfo.cluster);
			file_cache(mp3_fd, 0, fileinfo.size);
			state = S_CACHING;
			caching_step = 0;
			r = SP;
			printfd("S_READ_DIR, mp3_fd=%hd %x\r\n", mp3_fd, r);
			break;

		  case S_CACHING:
			r = SP;
			printfd("S_CACHING, mp3_fd=%d %x\r\n", mp3_fd, r);
			r = file_cache_work(mp3_fd);
			/* TODO: check return value from file_cache_work */
			if (++caching_step > 6) {
				file_seek(mp3_fd, 0);
				state = S_PLAYING_AND_CACHING;
				//debug = 0;
			}
			break;

		  case S_PLAYING_AND_CACHING:
			printfd("S_PLAYING_CACHING, mp3_fd=%d\r\n", mp3_fd);
			r = file_cache_work(mp3_fd);
			if (r) {
				state = S_PLAYING;
				ide_sleep();
				free_fat_memory();
			}
			/* fall through into S_PLAYING's code */

		  case S_PLAYING:
			num = play_queue_avail();
			printfd("S_PLAYING, num=%d\r\n", num);
			for (i=0; i<num; i++) {
				block = file_read_block(mp3_fd);
				printfd("FRB, r=%x\r\n", block);
				if (block == 0) break;  // not in cache yet
				if (block != 0xFFFF) {
					play_block(block, 4096);
				} else {
					state = S_FINISH_PLAYING;
					break;
				}
			}
			num = play_queue_avail();
			printfd("S_PLAYING(end), num=%d\r\n", num);
			break;

		  case S_FINISH_PLAYING:
			num = play_queue_avail();
			printfd("S_FINISH_PLAYING, mp3_fd=%d, num=%d\r\n",
			   mp3_fd, num);
			if (is_play_queue_empty() && is_ide_idle()) {
				file_close(mp3_fd);
				free_fat_memory();
				print_memory_available();
				ide_wakeup();
				state = S_READ_DIR;
			}
			break;

		  case S_PAUSED:
			if (prior_state == S_PLAYING_AND_CACHING) {
				r = file_cache_work(mp3_fd);
				if (r) {
					prior_state = S_PLAYING;
					ide_sleep();
					free_fat_memory();
				}
			}
			break;

		  default:
			printf("ILLEGAL STATE!!!!!\r\n");

		}
	}

	pm2_exit();
}



void read_directory(char dir, xdata struct dirent_struct *fileinfo)
{
	char r;
	unsigned char num=0;

	if (rand8_seed) {
		num = rand8();
		if (++rand_call_count > 10) {
			/* backup the random seed to NV memory every */
			/* 10th time we change it, so that the user */
			/* won't get the same "random" order every time */
			ibuf[0] = rand8_seed;
			write_flash_param(PARM_RAND8_SEED, ibuf);
			// printf("backing up rand8_seed = %d\r\n", attn);
			rand_call_count = 0;
		}
	}

	while (1) {
		if (rand8_seed) num--;
		if (num == 0) {
			r = dir_read(dir, fileinfo);
		} else {
			r = dir_read(dir, (void *)0);
		}
		// printf("dir_read, num=%d, r=%d\r\n", num, r);
		dir_position++;
		if (r) {
			file_seek(dir, 0);
			dir_position = 0;
			continue;
		}
		if (num > 0) continue;
		if (fileinfo->attrib & 0x10) {
			//printf("Skipping Directory: %s\r\n",
			//   fileinfo->long_name);
			continue;
		}
		if (!is_mp3_file(fileinfo)) {
			continue;
		}
		return;
	}
}


char is_mp3_file(xdata struct dirent_struct *fileinfo)
{
	if ((fileinfo->attrib & 0x1E) != 0) return 0;
	if (fileinfo->short_name[9] != 'M') return 0;
	if (fileinfo->short_name[10] != 'P') return 0;
	if (fileinfo->short_name[11] != '3') return 0;
	return 1;
}




char is_state_normal(unsigned char state)
{
	if (state == S_PLAYING
	 || state == S_PLAYING_AND_CACHING
	 || state == S_FINISH_PLAYING) {
		return 1;
	}
	return 0;
}


void main_directory_backup(char fd)
{
	unsigned int i;

	file_seek(fd, 0);
	if (dir_position <= 2) {
		dir_position = 0;
		return;
	}
	for (i=2; i<dir_position; i++) {
		dir_read(fd, (void *)0);
	}
	dir_position -= 2;
}

unsigned char rand8(void)
{
	_asm
	mov	dptr, #_rand8_seed
        movx    a, @dptr
        jnz     rand8b
        cpl     a
        movx    @dptr, a
rand8b: anl     a, #0xB8
        mov     c, p
        movx    a, @dptr
        rlc     a
        movx    @dptr, a
	mov	dpl, a
	_endasm;
}



